@@ -92,7 +92,7 @@ class RIMEImageBuilder:
                         if verbose:
                             print("reconnecting to the RIME backend...")
                         continue
-                    raise ValueError(e)
+                    raise ValueError(e.details()) from None
                 if verbose:
                     print("Status: {}".format(ManagedImage.Status.Name(image.status)))
                 if wait_until_finish:
