@@ -1,17 +1,18 @@
 """Library providing a module for file services."""
 from datetime import datetime, timezone
 from pathlib import Path
-from typing import List, Text
+from typing import Iterator, List, Optional, Text
 
-import grpc
 import requests
 from requests.adapters import HTTPAdapter
 from tqdm import tqdm
 from tqdm.utils import CallbackIOWrapper
 
+from rime_sdk.internal.grpc_error_handler import GRPCErrorHandler
 from rime_sdk.protos.file_upload.file_upload_pb2 import (
     GetDatasetFileUploadURLRequest,
     GetModelDirectoryUploadURLsRequest,
+    ListUploadedFileURLsRequest,
 )
 from rime_sdk.protos.file_upload.file_upload_pb2_grpc import FileUploadStub
 
@@ -21,6 +22,19 @@ def _now_str() -> str:
     return datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S%z")
 
 
+def _is_hidden(path: Path, dir_path: Path) -> bool:
+    return any(
+        part for part in path.relative_to(dir_path).parts if part.startswith(".")
+    )
+
+
+def _get_files_to_upload(dir_path: Path, upload_hidden: bool) -> Iterator[Path]:
+    paths: Iterator[Path] = dir_path.rglob("*")
+    if not upload_hidden:
+        paths = (path for path in paths if not _is_hidden(path, dir_path))
+    return paths
+
+
 class FileUploadModule:
     """A module that implements file uploading using a FileUploadStub."""
 
@@ -28,7 +42,9 @@ class FileUploadModule:
         """Create a FileUploadModule for the `file_upload_client`."""
         self._file_upload_client = file_upload_client
 
-    def upload_dataset_file(self, file_path: Path) -> str:
+    def upload_dataset_file(
+        self, file_path: Path, upload_path: Optional[str] = None,
+    ) -> str:
         """Upload dataset file ``file_path`` to RIME's blobstore via a FileUploadStub.
 
         The uploaded file is placed within the blobstore using its file name.
@@ -36,6 +52,9 @@ class FileUploadModule:
         Args:
             file_path: Path
                 Path to the file to be uploaded to RIME's blob store.
+            upload_path: Optional[str] = None
+                Name of the directory in the blob store file system. If omitted,
+                a unique random string will be the direcory.
 
         Returns:
             A reference to the uploaded file's location in the blob store. This
@@ -59,18 +78,20 @@ class FileUploadModule:
 
         file_name = file_path.name
         req = GetDatasetFileUploadURLRequest(file_name=file_name)
-        try:
+        if upload_path is not None:
+            req.upload_path = upload_path
+        with GRPCErrorHandler():
             resp = self._file_upload_client.GetDatasetFileUploadURL(request=req)
             self._impose_size_constraint(file_size, resp.upload_limit)
-        except grpc.RpcError as e:
-            # TODO(blaine): differentiate on different error types.
-            raise ValueError(e)
         self._upload_object(file_path, resp.upload_url)
         self._upload_string(f"Upload complete: {_now_str()}", resp.done_file_upload_url)
         return resp.destination_url
 
     def upload_model_directory(
-        self, dir_path: Path, upload_hidden: bool = False
+        self,
+        dir_path: Path,
+        upload_hidden: bool = False,
+        upload_path: Optional[str] = None,
     ) -> str:
         """Upload model directory ``dir_path`` to RIME's blobstore via a FileUploadStub.
 
@@ -85,6 +106,9 @@ class FileUploadModule:
             upload_hidden: bool = False
                 Whether or not to upload hidden files or subdirectories
                 (ie. those beginning with a '.').
+            upload_path: Optional[str] = None
+                Name of the directory in the blob store file system. If omitted,
+                a unique random string will be the direcory.
 
         Returns:
             A reference to the uploaded directory's location in the blob store. This
@@ -103,16 +127,7 @@ class FileUploadModule:
             raise FileNotFoundError(f"path '{dir_path}' does not exist")
         if not dir_path.is_dir():
             raise IOError(f"path '{dir_path}' is not a directory")
-        sub_paths = (
-            dir_path.rglob("*")
-            if upload_hidden
-            else filter(
-                lambda path: not any(
-                    (part for part in path.parts if part.startswith("."))
-                ),
-                dir_path.rglob("*"),
-            )
-        )
+        sub_paths = _get_files_to_upload(dir_path, upload_hidden)
         rel_paths: List[Text] = []
         total_size: int = 0
         for file_path in sub_paths:
@@ -122,20 +137,26 @@ class FileUploadModule:
         if len(rel_paths) == 0:
             raise IOError(f"directory '{dir_path}' is empty")
         req = GetModelDirectoryUploadURLsRequest(
-            directory_name=dir_path.name, relative_file_paths=rel_paths
+            directory_name=dir_path.name, relative_file_paths=rel_paths,
         )
-        try:
+        if upload_path is not None:
+            req.upload_path = upload_path
+        with GRPCErrorHandler():
             resp = self._file_upload_client.GetModelDirectoryUploadURLs(request=req)
             self._impose_size_constraint(total_size, resp.upload_limit)
-        except grpc.RpcError as e:
-            # TODO(blaine): differentiate on different error types.
-            raise ValueError(e)
         for rel_path in resp.upload_path_map:
             file_path = dir_path / rel_path
             self._upload_object(file_path, resp.upload_path_map[rel_path])
         self._upload_string(f"Upload complete: {_now_str()}", resp.done_file_upload_url)
         return resp.destination_url
 
+    def list_uploaded_files_urls(self) -> Iterator[str]:
+        """Return an iterator of file paths that have been uploaded."""
+        req = ListUploadedFileURLsRequest()
+        with GRPCErrorHandler():
+            resp = self._file_upload_client.ListUploadedFileURLs(request=req)
+        return resp.file_urls
+
     def _upload_object(self, file_path: Path, upload_url: str) -> None:
         """Upload the `file_path` to the location at `upload_url` via a PUT request."""
         file_size = file_path.stat().st_size
